import http.cookiejar
import os
import re
import sys
import urllib.request, urllib.parse, urllib.error
import subprocess
from tempfile import TemporaryDirectory

import lxml
from bs4 import BeautifulSoup

from crunchy.decoder import CrunchyDecoder
from crunchy.common import move_ask_overwrite, USER_AGENT


class CrunchyDownloader(object):

    def __init__(self, config_dict):
        for key, value in config_dict.items():
            setattr(self, key, value)

    def _get_player_revision(self, url):
        html = self._get_html(url)
        file_path = os.path.join(self.config_path, 'last_player_revision.txt')
        
        try:
            # As fallback, first try to use the last good known revision
            with open(file_path, 'r') as last_player_revision:
                fallback_player_revision = last_player_revision.readline().rstrip()
        except IOError:
            # A second fallback is a harcoded string with the last time I updated this file.
            # But in these cases you should use a proxy like Tor, since this is probably
            # a media region restriction
            fallback_player_revision = '20140613162835.83c7028de16b1d82f5e2c0b5e04405dd'

        with open(file_path, 'w') as last_player_revision:
            try:
                player_revision = re.findall(r'flash\\/(.+)\\/StandardVideoPlayer.swf', html).pop()
            except IndexError:
                url = url+'?skip_wall=1'  # Perv
                html = self._get_html(url)
                try:
                    player_revision = re.findall(r'flash\\/(.+)\\/StandardVideoPlayer.swf', html).pop()
                except IndexError:
                    player_revision = fallback_player_revision
            # Always write the last good know revision on a file, so we can load it later as a fallback
            last_player_revision.write(player_revision)

        return player_revision

    def _get_html(self, url):
        opener = urllib.request.build_opener(urllib.request.ProxyHandler({self.proxy_type: self.proxy}))
        opener.addheaders = [('Referer', 'http://crunchyroll.com/'), ('Host', 'www.crunchyroll.com'),
                             ('User-Agent', USER_AGENT)]
        res = opener.open(url).read()
        return res.decode(encoding='UTF-8')

    def _get_xml(self, req, media_id):
        url = 'http://www.crunchyroll.com/xml/'
        if req == 'RpcApiSubtitle_GetXml':
            data = {'req': 'RpcApiSubtitle_GetXml', 'subtitle_script_id': media_id}
        elif req == 'RpcApiVideoPlayer_GetStandardConfig':
            data = {'req': 'RpcApiVideoPlayer_GetStandardConfig', 'media_id': media_id,
                    'video_format': self.video_format, 'video_quality': self.resolution, 'auto_play': '1',
                    'show_pop_out_controls': '1', 'current_page': 'http://www.crunchyroll.com/'}
        else:
            data = {'req': req, 'media_id': media_id, 'video_format': self.video_format,
                    'video_encode_quality': self.resolution}
        cookie_jar = http.cookiejar.MozillaCookieJar(self.cookies_path)
        try:
            cookie_jar.load()
        except http.cookiejar.LoadError:
            sys.exit("Invalid 'cookies.txt' file. Run '{} -l' to remake 'cookies.txt' file.".format(sys.argv[0]))
        except IOError:
            sys.exit("Inexistent 'cookies.txt' file. Did you run '{} -l' first?".format(sys.argv[0]))
        cookie = urllib.request.HTTPCookieProcessor(cookie_jar)
        opener = urllib.request.build_opener(urllib.request.ProxyHandler({self.proxy_type: self.proxy}), cookie)
        opener.addheaders = [('Referer', 'http://static.ak.crunchyroll.com/flash/' +
                              self.player_revision + '/StandardVideoPlayer.swf'),
                             ('Host', 'www.crunchyroll.com'),
                             ('Content-type', 'application/x-www-form-urlencoded'),
                             ('User-Agent', USER_AGENT)]
        req = urllib.request.Request(url, urllib.parse.urlencode(data).encode(encoding='UTF-8'))
        res = opener.open(req).read().decode(encoding='UTF-8')
        return res

    def _get_video_url(self, url):
        if url.endswith('/'):
            url = url[:-1]
        res = self._get_html(url)
        series_url = urllib.parse.urlsplit(url).path
        regex = '<a href="({}/\S+)"\s+title="(.+)"\s+'.format(series_url)
        episodes = re.findall(regex, res)[::-1]
        print('List of episodes (not necessary in the correct order):')
        for i, episode_name in enumerate(episodes, 1):
            print('{}: {}'.format(i, episode_name[1]))
        while True:
            try:
                epnum = int(input('Episode number: '))
                if epnum <= 0:
                    raise IndexError
                episode_url = episodes[epnum-1][0]
                break
            except ValueError:
                print('Invalid value. Please enter a valid episode number!')
            except IndexError:
                print('Episode {} doesn\'t exist!'.format(epnum))
        url = urllib.parse.urljoin(url+'/', episode_url)
        print("Downloading video from URL '{}'".format(url))
        return url

    def get_video(self, page_url, subtitles_only=False):
        # http://www.crunchyroll.com/miss-monochrome-the-animation/episode-2-645085
        # page_url = 'http://www.crunchyroll.com/media-645085'
        if page_url.startswith('www'):
            page_url = 'http://'+page_url
        try:
            int(page_url)
            page_url = 'http://www.crunchyroll.com/media-'+page_url
        except ValueError:
            try:
                int(page_url[-6:])
            except ValueError:
                page_url = self._get_video_url(page_url)
        self.player_revision = self._get_player_revision(page_url)
        media_id = page_url[-6:]
        xmlconfig = BeautifulSoup(self._get_xml('RpcApiVideoPlayer_GetStandardConfig', media_id), 'xml')
        # xmlmeta = BeautifulSoup(self._get_xml('RpcApiVideoPlayer_GetMediaMetadata', media_id), 'xml')
        if '<code>4</code>' in xmlconfig:  # This is in VideoEncode_GetStreamInfo, but better to nip it in the bud early
            print('Video not available in your region.')
            sys.exit()
        vid_id = xmlconfig.find('media_id').string

        title = re.findall('<title>(.+?)</title>', self._get_html(page_url)).pop().replace('Crunchyroll - Watch ', '')
        title = title.replace('/', ' - ').replace(':', '-').replace('?', '.').replace('"', '\'').strip()

        # Normally 'RpcApiVideoEncode_GetStreamInfo' but some episodes f*ck up and show 1080p no matter the settings
        # xmlstream = BeautifulSoup(self._get_xml('RpcApiVideoPlayer_GetStandardConfig', media_id), 'xml')
        try:
            host = xmlconfig.find('host').string
        except AttributeError:
            print('Downloading 2 minute preview.')
            # xmlmeta = BeautifulSoup(self._get_xml('RpcApiVideoPlayer_GetMediaMetadata', media_id), 'xml')
            media_id = xmlconfig.find('media_id').string
            xmlconfig = BeautifulSoup(self._get_xml('RpcApiVideoEncode_GetStreamInfo', media_id), 'xml')
            try:
                host = xmlconfig.find('host').string
            except AttributeError:
                sys.exit(xmlconfig.find('msg').string)

        host_grr = re.search('fplive\.net', host)  # There was a time when fplive videos couldn't be downloaded, so...
        if host_grr:
            url1 = re.findall('.+/c[0-9]+', host).pop()
            url2 = re.findall('c[0-9]+\?.+', host).pop()
        else:
            url1 = re.findall('.+/ondemand/', host).pop()
            url2 = re.findall('ondemand/.+', host).pop()
        filename = xmlconfig.find('file').string

        xmllist = self._get_xml('RpcApiSubtitle_GetListing', media_id)
        xmllist = xmllist.replace('><', '>\n<')

        if '<media_id>None</media_id>' in xmllist:
            print('The video has hardcoded subtitles.')
            hardcoded = True
        else:
            try:
                regex = "id=([0-9]+)' title='.+{}.*'".format(self.lang.replace('(', '\(').replace(')', '\)'))
                sub_id = re.findall(regex, xmllist).pop()
                hardcoded = False
            except IndexError:
                try:
                    sub_id = re.findall("id\=([0-9]+)' title='.+English", xmllist).pop()  # Default back to English
                    print('Language not found, reverting to English')
                    hardcoded = False
                except IndexError:
                    print('The video\'s subtitles cannot be found, or are region-locked.')
                    hardcoded = True

        with TemporaryDirectory() as tmpdir:
            if not hardcoded:
                xmlsub = self._get_xml('RpcApiSubtitle_GetXml', sub_id)
                formattedSubs = CrunchyDecoder().return_subs(xmlsub)
                try:
                    with open(os.path.join(tmpdir, title+'.ass'), 'wb') as subfile:
                        subfile.write(formattedSubs.encode('utf-8-sig'))
                except IOError:
                    title = title.split(' - ', 1)[0]  # Episode name too long, splitting after episode number
                    with open(os.path.join(tmpdir, title+'.ass'), 'wb') as subfile:
                        subfile.write(formattedSubs.encode('utf-8-sig'))
                move_ask_overwrite(subfile.name, os.path.join(self.result_path, title+'.ass'))
            print("Subtitles for '{}' have been downloaded".format(title))
            # Exit this function if user asked only for subtitles.
            if subtitles_only:
                return None

            print('Downloading video...')
            video_path = os.path.join(tmpdir, title+'.flv')
            cmd = [self.rtmpdump_path, '-r', url1, '-a', url2, '-f', 'WIN 11,8,800,50', '-m', '15', '-W',
                   'http://static.ak.crunchyroll.com/flash/'+self.player_revision+'/ChromelessPlayerApp.swf',
                   '-p', page_url, '-y', filename, '-o', video_path]

            for i in range(self.retry+1):
                try:
                    subprocess.check_call(cmd)
                    move_ask_overwrite(video_path, os.path.join(self.result_path, title+'.flv'))
                    print("Video '{}' has been downloaded".format(title))
                    break
                except subprocess.CalledProcessError:
                    if i < self.retry:
                        print('Video failed to download, trying again. ({}/{})'.format(i+1, self.retry))
                    else:
                        sys.exit('Video failed to download too many times. Giving up...')
